package br.shmem.compiler;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import org.jdom.*;
import org.jdom.input.*;

/**
 *  Description of the Class
 *
 *@author     tminc
 *@created    July 2, 2002
 */
public class Compiler {
	private Document document;


	/**
	 *  Constructor for the Compiler object
	 *
	 *@param  document  Description of the Parameter
	 */
	public Compiler(Document document) {
		this.document = document;
	}


	/**
	 *  Gets the displayName attribute of the Compiler class
	 *
	 *@param  c  Description of the Parameter
	 *@return    The displayName value
	 */
	public static String getDisplayName(Class c) {
		if (c.isPrimitive()) {
			return c.toString();
		}
		else if (c.isArray()) {
			return getDisplayName(c.getComponentType()) + "[]";
		}
		else {
			return c.getName();
		}
	}


	/**
	 *  Gets the internalName attribute of the Compiler class
	 *
	 *@param  c  Description of the Parameter
	 *@return    The internalName value
	 */
	public static String getInternalName(Class c) {
		if (c.isPrimitive()) {
			if (c == Boolean.TYPE) {
				return "Boolean.TYPE";
			}
			else if (c == Character.TYPE) {
				return "Character.TYPE";
			}
			else if (c == Byte.TYPE) {
				return "Byte.TYPE";
			}
			else if (c == Short.TYPE) {
				return "Short.TYPE";
			}
			else if (c == Integer.TYPE) {
				return "Integer.TYPE";
			}
			else if (c == Long.TYPE) {
				return "Long.TYPE";
			}
			else if (c == Float.TYPE) {
				return "Float.TYPE";
			}
			else if (c == Double.TYPE) {
				return "Double.TYPE";
			}
			else {
				return "Void.TYPE";
			}
			// never reaches here
		}
		else if (c.isArray()) {
			return getDisplayName(c.getComponentType()) + "[].class";
		}
		else {
			return c.getName() + ".class";
		}
	}


	/**
	 *  Gets the parameter attribute of the Compiler class
	 *
	 *@param  c          Description of the Parameter
	 *@param  paramName  Description of the Parameter
	 *@return            The parameter value
	 */
	public static String getParameter(Class c, String paramName) {
		if (c.isPrimitive()) {
			if (c == Boolean.TYPE) {
				return "new Boolean(" + paramName + ")";
			}
			else if (c == Character.TYPE) {
				return "new Character(" + paramName + ")";
			}
			else if (c == Byte.TYPE) {
				return "new Byte(" + paramName + ")";
			}
			else if (c == Short.TYPE) {
				return "new Short(" + paramName + ")";
			}
			else if (c == Integer.TYPE) {
				return "new Integer(" + paramName + ")";
			}
			else if (c == Long.TYPE) {
				return "new Long(" + paramName + ")";
			}
			else if (c == Float.TYPE) {
				return "new Float(" + paramName + ")";
			}
			else if (c == Double.TYPE) {
				return "new Double(" + paramName + ")";
			}
			else {
				return "Void.TYPE";
			}
			// never reaches here
		}
		else {
			return paramName;
		}
	}


	public static String getReturn(Class c, String resultName) {
		if (c.isPrimitive()) {
			if (c == Boolean.TYPE) {
				return "((Boolean) " + resultName + ").booleanValue()";
			}
			else if (c == Character.TYPE) {
				return "((Character) " + resultName + ").charValue()";
			}
			else if (c == Byte.TYPE) {
				return "((Byte)" + resultName + ").byteValue()";
			}
			else if (c == Short.TYPE) {
				return "((Short)" + resultName + ").shortValue()";
			}
			else if (c == Integer.TYPE) {
				return "((Integer)" + resultName + ").intValue()";
			}
			else if (c == Long.TYPE) {
				return "((Long)" + resultName + ").longValue()";
			}
			else if (c == Float.TYPE) {
				return "((Float)" + resultName + ").floatValue()";
			}
			else if (c == Double.TYPE) {
				return "((Double)" + resultName + ").doubleValue()";
			}
			else {
				return "";
				// never reaches here
			}
		}
		else {
			return "(" + getDisplayName(c) + ") " + resultName;
		}
	}

	/**
	 *  Description of the Method
	 *
	 *@param  args  Description of the Parameter
	 */
	public static void help(String args[]) {
		boolean show = false;

		if (args.length < 2) {
			show = true;
		}
		for (int i = args.length - 1; !show && i >= 0; i--) {
			if (args[i].startsWith("--help")) {
				show = true;
			}
		}
		help(show);
	}


	/**
	 *  Description of the Method
	 *
	 *@param  show  Description of the Parameter
	 */
	public static void help(boolean show) {
		if (show) {
			System.err.println("Shmem Shared Reference compiler.");
			System.err.println("Arguments:");
			System.err.println("\tclass file: a .class file containing a Java class");
			System.err.println("\txml file: a .xml file describing default parameters");
			System.err.println("Examples:");
			System.err.println("\tshmemc classes/MyClass.class MyClass.xml");
			System.err.println("\t\tloads the file MyClass.class in the classes subdirectory, with the xml file MyClass.xml");
			System.err.println("\tshmemc MyClass.class -");
			System.err.println("\t\tparses the xml information from stdin");
			System.exit(1);
		}
	}


	/**
	 *  Description of the Method
	 *
	 *@param  args                       Description of the Parameter
	 *@return                            Description of the Return Value
	 *@exception  FileNotFoundException  Description of the Exception
	 */
	public static FileInputStream loadClassFile(String args[])
		throws FileNotFoundException {
		int idx = 0;
		// start from 0
		while (idx < args.length) {
			if (args[idx].endsWith(".class")) {
				System.out.println("Loading input class file " + args[idx] + "...");
				return new FileInputStream(args[idx]);
			}
			idx++;
		}
		return null;
	}


	/**
	 *  Description of the Method
	 *
	 *@param  args                       Description of the Parameter
	 *@return                            Description of the Return Value
	 *@exception  FileNotFoundException  Description of the Exception
	 */
	public static InputStream loadXMLFile(String args[])
		throws FileNotFoundException {
		int idx = 1;
		// start from 1
		while (idx < args.length) {
			if (args[idx].endsWith(".xml")) {
				System.out.println("Loading input XML file " + args[idx] + "...");
				return new FileInputStream(args[idx]);
			}
			else if (args[idx].equals("-")) {
				return System.in;
			}
			idx++;
		}
		return null;
	}


	/**
	 *  Description of the Method
	 *
	 *@param  args  Description of the Parameter
	 */
	public final static void main(String args[]) {
		InputStream xmlFile;
		FileInputStream classFile;

		help(args);

		try {
			classFile = loadClassFile(args);
			xmlFile = loadXMLFile(args);

			SAXBuilder builder = new SAXBuilder(true);
			ShmemClassLoader loader = new ShmemClassLoader(classFile);
			Document document = builder.build(xmlFile, "./");
			Compiler compiler = new Compiler(document);

			compiler.writeClass(loader.getLoadedClass());
		} catch (Exception e) {
			e.printStackTrace();
		}
	}


	/**
	 *  Description of the Method
	 *
	 *@param  str  Description of the Parameter
	 *@return      Description of the Return Value
	 */
	public static String removeSpaces(String str) {
		StringBuffer strBuf = new StringBuffer(str.length());

		for (int i = 0; i < str.length(); i++) {
			char c = str.charAt(i);

			if (c != ' ') {
				strBuf.append(c);
			}
		}
		return strBuf.toString();
	}


	/**
	 *  Description of the Method
	 *
	 *@param  out  Description of the Parameter
	 *@param  c    Description of the Parameter
	 */
	public void writeClass(java.io.PrintStream out, Class c) 
		throws NotSerializableException 
	{
		if (!java.io.Serializable.class.isAssignableFrom(c))
			throw new NotSerializableException(c);
		String referenceName = document.getRootElement().getAttributeValue("name");
		int split_idx = referenceName.lastIndexOf('.');
		String cName = referenceName;

		if (split_idx > 0) {
			String pkgName = referenceName.substring(0, split_idx);
		 
			out.println("package " + pkgName + ";");
			out.println();
			cName = referenceName.substring(split_idx + 1, referenceName.length());
		}

		out.println();
		out.println("import br.shmem.*;");
		out.println("import java.lang.reflect.*;");
		out.println();

		String tail = " {";

		if (c.isInterface()) {
			tail = " implements " + c.getName() + tail;
		}

		out.println(Modifier.toString(c.getModifiers()) + " class " + cName +
				" extends SharedReference" + tail);

		out.println("public " + cName + "(SharedMemory shmem, String globalName) {");
		out.println("  super(shmem, globalName);");
		out.println("}");

		out.println();

		Method[] methods = c.getDeclaredMethods();

		for (int i = 0; i < methods.length; i++) {
			try {
				List l = document.getRootElement().getChildren("method");
				Iterator it = l.iterator();
				Element el = null;

				while (it.hasNext() && el == null) {
					Element e = (Element) it.next();

					if (methods[i].toString().equals(e.getAttributeValue("name"))) {
						el = e;
					}
				}
				writeMethod(out, methods[i], el);
				out.println();
			} catch (InvalidModifierException e) {
				e.printStackTrace(System.err);
			}
		}
		out.println("}");
	}


	/**
	 *  Description of the Method
	 *
	 *@param  c                                  Description of the Parameter
	 *@exception  java.io.FileNotFoundException  Description of the Exception
	 */
	public void writeClass(Class c)
		throws java.io.FileNotFoundException,
		NotSerializableException
		{
		String fileName = document.getRootElement().getAttributeValue("name");

		fileName = fileName.replace('.', File.separatorChar) + ".java";

		File dest = new File(fileName);

		if(dest.getParentFile() != null)
		    dest.getParentFile().mkdirs();

		writeClass(new PrintStream(new FileOutputStream(dest)), c);
		System.out.println("Java source written to " + fileName + " .");
	}


	/**
	 *  Description of the Method
	 *
	 *@param  out                           Description of the Parameter
	 *@param  method                        Description of the Parameter
	 *@param  el                            Description of the Parameter
	 *@exception  InvalidModifierException  Description of the Exception
	 */
	public void writeMethod(java.io.PrintStream out, Method method, Element el)
		throws InvalidModifierException {
		boolean write = false;

		if (el != null) {
			System.out.println("Element: " + el);
			System.out.println("\t" + el.getAttribute("write"));
			write = new Boolean(el.getAttributeValue("write")).booleanValue();
		}

		int modifiers = method.getModifiers();

		if (Modifier.isPrivate(modifiers)) {
			return;
		}

		if (Modifier.isProtected(modifiers)) {
			return;
		}

		if (Modifier.isStatic(modifiers)) {
			return;
		}

		if (Modifier.isAbstract(modifiers)) {
			throw new InvalidModifierException("The method can not be abstract.");
		}

		if (Modifier.isNative(modifiers)) {
			throw new InvalidModifierException("The method can not be native.");
		}

		out.print("  ");
		if (Modifier.isPublic(modifiers)) {
			out.print("public ");
		}

		if (Modifier.isFinal(modifiers)) {
			out.print("final ");
		}
		
		out.print(getDisplayName(method.getReturnType()));

		try {
			br.shmem.SharedReference.class.getMethod(method.getName(), method.getParameterTypes());
			out.print(" " + method.getName() + "Proxy(");
		} catch (NoSuchMethodException e) {
			out.print(" " + method.getName() + "(");
		}

		Class[] params = method.getParameterTypes();

		if (params.length == 0) {
			out.println(")");
		}

		for (int i = 0; i < params.length; i++) {
			out.print(getDisplayName(params[i]) + " p" + (i + 1));
			if (i + 1 == params.length) {
				out.println(")");
			}
			else {
				out.print(", ");
			}
		}

		// inserir throws clause
		out.print("    throws ShmemException");

		Class[] exceptions = method.getExceptionTypes();

		for (int i = 0; i < exceptions.length; i++) {
			if (br.shmem.ShmemException.class.isAssignableFrom(exceptions[i])) continue;
			out.println(",");
			out.print("           " + getDisplayName(exceptions[i]));
		}

		out.println(" {");
		out.println("    try {");
		if (write) {
			out.println("      beginTrans();");
		}

		// argument types:
		if (params.length == 0) {
			out.println("      Class[] argTypes = null;");
		}
		else {
			out.print("      Class[] argTypes = new Class[] {");
			for (int i = 0; i < params.length; i++) {
				out.print("                                      " + getInternalName(params[i]));
				if (i + 1 == params.length) {
					out.println("};");
				}
				else {
					out.println(",");
				}
			}
		}

		// argument values:
		if (params.length == 0) {
			out.println("      Object[] args = null;");
		}
		else {
			out.print("      Object[] args = new Object[] {");

			for (int i = 0; i < params.length; i++) {
				out.print("                                    " + getParameter(params[i], "p" + (i + 1)));
				if (i + 1 == params.length) {
					out.println("};");
				}
				else {
					out.println(",");
				}
			}
		}

		out.println();
		out.println("      Object result = callMethod(\"" + method.getName() + "\", argTypes, args);");

		if (write) {
			out.println("      commit();");
		}
		
		if (method.getReturnType() != Void.TYPE) {
			out.println("      return " + getReturn( method.getReturnType(), "result") + ";"); 
		}

		out.println("    } catch (CommLayerException e) {");
		out.println("      throw new ShmemException(e);");
		out.println("    } catch (InvocationException e) {");
		out.println("      throw new ShmemException(e);");
		out.println("    } catch (InsecureObjectException e) {");
		out.println("      throw new ShmemException(e);");
		out.println("    } catch (InvalidOperationException e) {");
		out.println("      throw new ShmemException(e);");
		out.println("    } catch (InvocationTargetException e) {");
		// downcast throwables:
		for (int i = 0; i < exceptions.length; i++) {
			String eName = exceptions[i].getName();

			out.print("      ");
			if (i > 0) {
				out.print("else ");
			}
			out.println("if (e.getTargetException() instanceof " + eName + ") {");
			out.println("        throw (" + eName + ") e.getTargetException();");
			out.println("      }");
		}
		// failsafe code to throw an unknown exception:
		out.println("      throw new ShmemException(e);");
		out.println("    }");

		out.println("  }");

	}
}

