package wumpus.agents;

import wumpus.*;
import java.util.Random;
import javax.swing.JComponent;
import java.awt.*;

public class WumpusAgent implements Runnable {

	public static final double rd_ChangeDirection = 0.4; // 50% chance to turn;
	public static final long actionFrequency = 8000; // milisseconds
	public static final Random lazyRandomizer = new Random();
	
	private boolean alive = true;
	private int currentDirection = Board.NORTH;
	private Walker delegate = null;
	private Board board;
	
	public WumpusAgent(Board board) {
		int x, y;
		this.board = board;
		board.setWumpusAgent(this);
		Square s;
		do {
			x = lazyRandomizer.nextInt(board.getWidth());
			y = lazyRandomizer.nextInt(board.getHeight());
			s = board.getSquare(x, y);
		} while (!s.isHole() && !s.isExit() && !s.isOccupied());
		delegate = new Walker(board, x, y);
		delegate.setRenderer(new WumpusRenderer());
		try {
			s.occupy(delegate);
		} catch (OccupationException e) {
			System.err.println("this should never happen!");
			e.printStackTrace(System.err);
		}
	}
	
	public void receiveArrow(int x, int y) {
		synchronized(delegate) {
			if (x == delegate.getX() &&
			y == delegate.getY()) {
				alive = false;
				board.getSquare(x, y).leave(delegate);
			}
		}
	}
	
	private int nextX() {
		if (currentDirection == Board.EAST)
			return delegate.getX() +1;
		else if (currentDirection == Board.WEST)
			return delegate.getX() - 1;
		else
			return delegate.getX();
	}
	
	private int nextY() {
		if (currentDirection == Board.SOUTH)
			return delegate.getY() +1;
		else if (currentDirection == Board.NORTH)
			return delegate.getY() - 1;
		else
			return delegate.getY();
	}
	
	protected void changeDirection(int lastDirection, int tries) {
		if (tries > 1) {
			currentDirection = (lastDirection + 1 + lazyRandomizer.nextInt(3)) % 4;
		}
		else { // if it's his last chance, turn around.
			currentDirection = (lastDirection + 2) % 4;
		}
	}

	protected void action() 
		throws OccupationException
	{
		int nextX = nextX();
		int nextY = nextY();
		int tries = 4;
		int lastDirection = currentDirection;
		// random direction change:
		if (lazyRandomizer.nextDouble() < rd_ChangeDirection) {
			changeDirection(lastDirection, 2);
			nextX = nextX();
			nextY = nextY();
		}
		Square sq = board.getSquare(nextX, nextY);
		while (tries > 0 && sq.isHole() || sq.isWall()) {
			changeDirection(lastDirection, tries--);
			nextX = nextX();
			nextY = nextY();
			sq = board.getSquare(nextX, nextY);
		}
		eat(sq);
		delegate.move(currentDirection);
		board.fireRepaintRequest();
	}
	
	protected void eat(Square sq) {
		synchronized(sq) {
			if (alive && sq.isOccupied()) {
				board.getGame().setDeath();
			}
		}
	}
	
	public void run() {
		try {
			Game g = board.getGame();
			synchronized(g) {
				if (!g.isGameStarted()) g.wait();
			}
			Object waitLock = new Object();
			while (alive && !g.isGameEnded()) {
				try {
					action();
					synchronized(waitLock) {
						waitLock.wait(actionFrequency);
					}
				} catch (OccupationException e) {
					System.err.println(e.toString() + ": (" + delegate.getX() + ", " + delegate.getY() + ")");
				}
			}
		} catch (InterruptedException e) {
			
		}
	}
	
	class WumpusRenderer extends JComponent {
		
		public void paintComponent(Graphics g) {
			int w = getWidth();
			int h = getHeight();
			g.setColor(Color.red);
			g.fillRect(w/4, h/4, w/2, h/2);
		}
	}
	
}
